package kc.mega.gun;

import kc.mega.BeepBoop;
import kc.mega.game.BotState;
import kc.mega.game.GameState;
import kc.mega.learning.Model;
import kc.mega.other.DatasetWriter;
import kc.mega.other.HitRateTracker;
import kc.mega.other.Strategy;
import kc.mega.utils.Geom;
import kc.mega.wave.WaveManager;
import kc.mega.wave.WaveWithFeatures;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import robocode.AdvancedRobot;
import robocode.BulletHitBulletEvent;
import robocode.BulletHitEvent;
import robocode.util.Utils;

public class Gun {
  private static boolean useVirtualGuns = false;

  private final AdvancedRobot bot;
  private final Strategy strategy;
  private final WaveManager.MyWaveManager waveManager;
  private final HitRateTracker hitRateTracker;
  private final DatasetWriter datasetWriter;
  private final List<Model> aimModels;
  private final List<Integer> aimModelHits;

  private boolean hasFired;
  private Model firingModel;
  private List<Double> lastAimAngles;

  public Gun(AdvancedRobot bot, Strategy strategy, WaveManager.MyWaveManager waveManager,
      HitRateTracker hitRateTracker, DatasetWriter datasetWriter) {
    this.bot = bot;
    this.strategy = strategy;
    this.waveManager = waveManager;
    this.hitRateTracker = hitRateTracker;
    this.datasetWriter = datasetWriter;
    aimModels = Arrays.asList(AimModels.getMainGun(), AimModels.getAntiSurferGun());
    aimModelHits = Arrays.asList(1, 0);
    firingModel = aimModels.get(0);
  }

  public void init() {
    hasFired = false;
    System.out.println("Using " + firingModel.name + " gun");
  }

  public void aim(GameState gs) {
    BotState myState = gs.getMyState();
    BotState enemyState = gs.getEnemyState();
    for (GunWave w : waveManager.updateWaves(enemyState)) {
      for (Model aimModel : aimModels) {
        aimModel.train(w);
      }
      if (w.hasBullet || (w.virtuality == 0 && w.didCollide)) {
        hitRateTracker.logShotPassed(w.power);
      }
      if (w.hasBullet && useVirtualGuns) {
        w.updateModelHits(aimModelHits);
      }
      if (BeepBoop.writeData) {
        datasetWriter.write("gun-data", w.asArray());
      }
    }

    double bulletPower = strategy.ram ? 0 : Math.min(firingModel.hasData() ?
        3 : Math.nextAfter(0.15, -1), BulletPowerSelector.bestBulletPower(gs));
    boolean doFire = false;
    if (bulletPower > 0.099999 &&
        myState.gunHeat <= 0.0 &&
        Math.abs(bot.getGunTurnRemainingRadians()) < 0.0001) {
      hasFired = true;
      doFire = true;
      gs.onMyFire(bulletPower);
      gs.bot.setFireBullet(bulletPower);
    }

    if (gs.myHistory.size() <= 2) {
      return;
    }

    Point2D.Double enemyLocation = enemyState.location;
    bot.setTurnGunRightRadians(Utils.normalRelativeAngle(Geom.absoluteBearing(
        gs.myFuture.get(0).location, enemyLocation) - bot.getGunHeadingRadians()));

    int ticksUntilFire = (int)Math.ceil(myState.gunHeat / bot.getGunCoolingRate());
    if (!doFire && ticksUntilFire <= 3) {
      List<BotState> myNextHistory = new ArrayList<>(gs.myHistory);
      List<BotState> enemyNextHistory = new ArrayList<>(gs.enemyHistory);
      myNextHistory.add(0, gs.myFuture.get(0));
      enemyNextHistory.add(0, gs.enemyFuture.get(0));
      WaveWithFeatures aimWave = (WaveWithFeatures)new WaveWithFeatures.Builder()
          .waveManager(waveManager)
          .myHistory(myNextHistory)
          .enemyHistory(enemyNextHistory)
          .power(bulletPower)
          .hasBullet(true)
          .build();
      double aimAngle = ticksUntilFire > 1 ? firingModel.getAimAngleFast(aimWave) :
        firingModel.getAimAngle(aimWave);
      bot.setTurnGunRightRadians(Utils.normalRelativeAngle(aimAngle - bot.getGunHeadingRadians()));
      if (ticksUntilFire <= 1 && useVirtualGuns) {
        lastAimAngles = new ArrayList<>();
        for (Model model : aimModels) {
          lastAimAngles.add(model == firingModel ? aimAngle : model.getAimAngle(aimWave));
        }
      }
    } else {
      pickFiringModel();
    }

    if (hasFired && bulletPower > 0.099999) {
      GunWave w = (GunWave)new GunWave.Builder()
          .gunHeading(bot.getGunHeadingRadians())
          .aimAngles(lastAimAngles)
          .waveManager(waveManager)
          .myHistory(gs.myHistory)
          .enemyHistory(gs.enemyHistory)
          .power(bulletPower)
          .hasBullet(doFire)
          .doPaint(BeepBoop.paint)
          .build();
      waveManager.add(w);
    }
  }

  public void pickFiringModel() {
    int bestModelInd = 0;
    if (useVirtualGuns) {
      for (int i = 1; i < aimModels.size(); i++) {
        if (aimModelHits.get(i) > aimModelHits.get(bestModelInd)) {
          bestModelInd = i;
        }
      }
    } else if (hitRateTracker.hitRateInBounds(0, 0.12)) {
      bestModelInd = 1;
    }
    if (aimModels.get(bestModelInd) != firingModel) {
      firingModel = aimModels.get(bestModelInd);
      System.out.println("Switching to " + firingModel.name + " gun");
    }
  }

  public void onBulletHit(BulletHitEvent e, GameState gs) {
    hitRateTracker.logHit(e.getBullet().getPower(), gs.getEnemyState().energy);
    GunWave hitWave = waveManager.getHitWave(e.getBullet().getPower());
    if (hitWave != null) {
      for (GunWave w : waveManager.getWaves()) {
        if (!w.hasBullet && Math.abs(w.fireTime - hitWave.fireTime) <= 3) {
          w.didHit = true;
        }
      }
    }
  }

  public void onBulletHitBullet(BulletHitBulletEvent e) {
    if (bot.getOthers() > 0) {
      GunWave hitWave = waveManager.getBulletHitBulletWave(
          new Point2D.Double(e.getBullet().getX(), e.getBullet().getY()),
          e.getBullet().getPower(), bot.getTime());
      if (hitWave != null) {
        for (GunWave w : waveManager.getWaves()) {
          if (!w.hasBullet && Math.abs(w.fireTime - hitWave.fireTime) <= 3) {
            w.didCollide = true;
          }
        }
      }
    }
  }

  public void printStats() {
    System.out.println("My Hit Rate: " + hitRateTracker.getHitRateStr());
    if (BeepBoop.TC) {
      System.out.println("TC Score: " + (float)(
          hitRateTracker.getDamage() / (1 + bot.getRoundNum())));
    }
    if (useVirtualGuns && BeepBoop.verbose) {
      System.out.println("Virtual gun stats");
      for (int i = 0; i < aimModels.size(); i++) {
        System.out.println(" " + aimModels.get(i).name + (aimModels.get(i) == firingModel ?
            "* " : " ") + aimModelHits.get(i));
      }
    }
  }
}
