package kc.mega.learning;

import kc.mega.utils.KUtils;
import kc.mega.wave.WaveWithFeatures;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import ags.utils.KdTree;
import ags.utils.KdTree.Entry;
import jk.math.FastTrig;

public abstract class KNNModel<T> extends Model {
  protected final KdTree<T> tree;
  protected final GFBins bins;
  protected final String[] features;
  protected final double[][] params;
  protected final double distanceScale;
  protected final int maxNeighbors;
  protected final int maxNeighborsFast;
  protected final double neighborhoodSizeDivider;

  public static abstract class Builder<T> {
    private final String name;
    private GFBins bins = null;
    private String[] features;
    private double[][] params;
    private double distanceScale;
    private double neighborhoodSizeDivider;
    private int maxNeighbors;
    private int maxNeighborsFast = 0;
    private int maxTreeSize = 50000;

    public Builder(String name) {
      this.name = name;
    }

    public abstract KNNModel<T> build();
    public Builder<T> bins(GFBins val) {bins = val; return this;}
    public Builder<T> features(String[] val) {features = val; return this;}
    public Builder<T> params(double[][] val) {params = val; return this;}
    public Builder<T> distanceScale(double val) {distanceScale = val; return this;}
    public Builder<T> neighborhoodSizeDivider(double val) {neighborhoodSizeDivider = val; return this;}
    public Builder<T> maxNeighbors(int val) {maxNeighbors = val; return this;}
    public Builder<T> maxNeighborsFast(int val) {maxNeighborsFast = val; return this;}
    public Builder<T> maxTreeSize(int val) {maxTreeSize = val; return this;}
  }

  public KNNModel(Builder<T> builder) {
    super(builder.name);
    tree = new KdTree.Manhattan<>(builder.features.length, builder.maxTreeSize);
    this.bins = builder.bins;
    this.features = builder.features;
    this.params = builder.params;
    this.distanceScale = builder.distanceScale;
    this.maxNeighbors = builder.maxNeighbors;
    this.maxNeighborsFast = builder.maxNeighborsFast;
    this.neighborhoodSizeDivider = builder.neighborhoodSizeDivider;
  }

  @Override
  public boolean hasData() {
    return tree.size() > 0;
  }

  protected int getNumNeighbors() {
    return Math.min(maxNeighbors, Math.max(5, (int)(tree.size() / neighborhoodSizeDivider)));
  }

  protected double[] embed(WaveWithFeatures w) {
    return embed(getNormalizedFeatures(w));
  }

  protected double[] embed(Map<String, Double> normalizedFeatures) {
    double[] embedding = new double[features.length];
    for (int i = 0; i < features.length; i++) {
      double featureVal = normalizedFeatures.get(features[i]);
      embedding[i] = params[i].length == 1 ? params[i][0] * featureVal :
        params[i][0] * Math.pow(1e-4 + params[i][1] + featureVal, params[i][2]);
    }
    return embedding;
  }

  protected double[] getWeights(List<Entry<T>> neighbors) {
    double[] weights = new double[neighbors.size()];
    double maxLogit = Double.NEGATIVE_INFINITY;
    for (int i = 0; i < neighbors.size(); i++) {
      double distance = neighbors.get(i).distance;
      double logit = distance * distanceScale;
      weights[i] = logit;
      maxLogit = Math.max(logit, maxLogit);
    }
    KUtils.softmax(weights, maxLogit);
    return weights;
  }

  protected static Map<String, Double> getNormalizedFeatures(WaveWithFeatures w) {
    double bft = w.distance / w.speed;
    double latVel = w.velocity * FastTrig.sin(w.relativeHeading);
    double advVel = w.velocity * FastTrig.cos(w.relativeHeading);
    double advDir = w.moveDirection * FastTrig.cos(w.relativeHeading);
    Map<String, Double> normalizedFeatures = new HashMap<>();
    normalizedFeatures.put("virtuality", w.virtuality / 5);
    normalizedFeatures.put("bft", bft / 100);
    normalizedFeatures.put("accel", Math.max(2 + w.accel, 0) / 2);
    normalizedFeatures.put("accelSign", Math.signum(w.accel));
    normalizedFeatures.put("latVel", Math.abs(latVel) / 8);
    normalizedFeatures.put("vel", Math.abs(w.velocity) / 8);
    normalizedFeatures.put("advVel", (advVel + 16) / 8);
    normalizedFeatures.put("advDir", (advDir + 1) / 2);
    normalizedFeatures.put("vChangeTimer", Math.min(w.vChangeTimer, 70) / bft);
    normalizedFeatures.put("dirChangeTimer", Math.min(w.dirChangeTimer, 70) / bft);
    normalizedFeatures.put("decelTimer", Math.min(w.decelTimer, 70) / bft);
    normalizedFeatures.put("distanceLast10", w.distanceLast10 / 80);
    normalizedFeatures.put("distanceLast20", w.distanceLast20 / 160);
    normalizedFeatures.put("orbitalWallAhead", w.orbitalWallAhead / 1.5);
    normalizedFeatures.put("orbitalWallReverse", w.orbitalWallReverse / 1.5);
    normalizedFeatures.put("maeWallAhead", w.maeWallAhead);
    normalizedFeatures.put("maeWallReverse", w.maeWallReverse);
    normalizedFeatures.put("stickWallAhead", w.stickWallAhead / (Math.PI / 2));
    normalizedFeatures.put("stickWallReverse", w.stickWallReverse / (Math.PI / 2));
    normalizedFeatures.put("gameTime", w.fireTime / 500.0);
    normalizedFeatures.put("shotsFired", w.shotsFired / 1000.0);
    normalizedFeatures.put("currentGF", (1 + w.currentGF) / 2);
    normalizedFeatures.put("didHit", w.didHit ? 1.0 : 0.0);
    normalizedFeatures.put("didCollide", w.didCollide ? 1.0 : 0.0);
    return normalizedFeatures;
  }
}
