package kc.mega.movement;

import kc.mega.learning.GFBins;
import kc.mega.learning.Model;
import kc.mega.learning.SimpleModels;
import kc.mega.other.HitRateTracker;
import kc.mega.other.Painter;
import kc.mega.other.Stopwatch;
import kc.mega.other.Strategy;
import kc.mega.utils.Geom;
import kc.mega.utils.KUtils;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jk.math.FastTrig;

public class DangerComputer {
  private static final int N_BINS = 151;
  private static final double HIT_ROLLING = 0.99;
  private static final double PRIOR_STRENGTH = 5;
  private static final double MIN_PRIOR_WEIGHT = 0.1;
  private static final double WEIGHT_POWER = 2;
  private static final double FORCE_ACTIVE_WEIGHT = 2;

  private final GFBins bins;
  private final Strategy strategy;
  private final HitRateTracker hitRateTracker;
  private final Map<MovementWave, WaveDangerTracker> waveDangerTrackers;
  private final List<DangerEstimator> activeEstimators;
  private final List<DangerEstimator> dangerEstimators;

  private boolean antiHOT = true;

  public DangerComputer(Strategy strategy, HitRateTracker hitRateTracker) {
    this.strategy = strategy;
    this.hitRateTracker = hitRateTracker;
    bins = new GFBins(N_BINS);
    waveDangerTrackers = new HashMap<>();
    activeEstimators = new ArrayList<>();
    dangerEstimators = new ArrayList<>();

    dangerEstimators.add(new DangerEstimator(new SimpleModels.HOTModel(bins), true, 0.0, 0.05, 1));
    dangerEstimators.add(new DangerEstimator(new SimpleModels.SimpleTargetingModel(
        "Linear", bins, false, false, false), true, 0.0, 0.05, 2));
    dangerEstimators.add(new DangerEstimator(new SimpleModels.SimpleTargetingModel(
        "AvgLinear", bins, false, false, true), true, 1, 1, 0));
    dangerEstimators.add(new DangerEstimator(new SimpleModels.SimpleTargetingModel(
        "Circular", bins, true, true, false), true, 0.0, 0.05, 2));
    dangerEstimators.add(new DangerEstimator(
        new SimpleModels.CurrentGFModel(bins), true, 1, 1, 0));

    dangerEstimators.add(new DangerEstimator(DangerModels.simple(bins), true, 0.0, 1, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.simple2(bins), true, 0.03, 1, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiThorn(bins), true, 0.05, 1, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiDruss(bins), true, 0.09, 1.0, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiDiamond(bins), true, 0.08, 1.0, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiWaveSerpent(bins), true, 0.08, 1.0, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiCassiusClay(bins), true, 0.08, 1.0, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiKomarious(bins), true, 0.05, 1, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiRaikoNano(bins), true, 0.01, 0.06, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiSplinter(bins), true, 0.03, 0.06, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiWaveShark(bins), true, 0.06, 0.08, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiWaylander(bins), true, 0.08, 1.0, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiMicroAspid(bins), true, 0.05, 0.07, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiGrinnik(bins), true, 0.03, 0.06, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiSedan(bins), true, 0.04, 0.08, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.antiChiva(bins), true, 0.05, 0.09, 10));

    dangerEstimators.add(new DangerEstimator(DangerModels.flattener1(bins), false, 0.09, 1, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.flattener2(bins), false, 0.08, 1, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.flattener3(bins), false, 0.08, 1, 10));
    dangerEstimators.add(new DangerEstimator(DangerModels.flattener4(bins), false, 0.08, 1, 10));
  }

  public void init() {
    waveDangerTrackers.clear();
    setActiveEstimators();
  }

  public GFBins getBins() {
    return bins;
  }

  public void remove(MovementWave w) {
    waveDangerTrackers.remove(w);
  }

  public void onNonHeadOnHit() {
    antiHOT = false;
    setActiveEstimators();
  }

  public void onChangedShadows(List<MovementWave> waves) {
    for (MovementWave w : waves) {
      if (waveDangerTrackers.containsKey(w)) {
        waveDangerTrackers.get(w).onUpdatedShadow();
      }
    }
  }

  public void learnFromHit(MovementWave w) {
    setActiveEstimators();
    if (!w.hasBullet) {   // currently only picking estimator weights from bullet collisions
      for (DangerEstimator estimator : dangerEstimators) {
        estimator.updateWeightingStats(w);
      }
    }

    for (DangerEstimator estimator : dangerEstimators) {
      if (estimator.isHitModel) {
        estimator.model.train(w);
        for (Map.Entry<MovementWave, WaveDangerTracker> e : waveDangerTrackers.entrySet()) {
          e.getValue().onUpdatedEstimator(estimator);
        }
      }
    }
  }

  public void learnFromVisit(MovementWave w, boolean enemyIsAlive) {
    setActiveEstimators();
    if (enemyIsAlive) {
      for (DangerEstimator estimator : dangerEstimators) {
        if (!estimator.isHitModel) {
          estimator.model.train(w);
        }
      }
    }
    remove(w);
  }

  public double getDanger(MovementWave w, double[] offsetRange) {
    double[] GFRange = w.getGFRange(offsetRange);
    double[] alreadyHitGFRange = w.aggHitOffsetRange[0] < 0 ? w.aggHitOffsetRange : null;
    double[] dangers = getDangerTracker(w).getDangers();
    double danger = 0;
    for (int bin = bins.hitBin(GFRange[0]); bin <= bins.hitBin(GFRange[1]); bin++) {
      double binWeight = bins.binWeight(GFRange, bin);
      // if part of the wave has already hit us, don't add danger from it
      if (alreadyHitGFRange != null && alreadyHitGFRange[0] < bins.upperGF[bin] &&
          alreadyHitGFRange[1] > bins.lowerGF[bin]) {
        binWeight *= 1 - bins.binWeight(alreadyHitGFRange, bin);
      }
      danger += dangers[bin] * binWeight;
    }
    return danger;
  }

  public double getApproximatePointDanger(MovementWave w, Point2D.Double location) {
    double halfWidth = FastTrig.atan(Geom.MAX_BOT_HALF_WIDTH / w.source.distance(location));
    halfWidth /= Math.min(w.maePrecise[0], w.maePrecise[1]);
    int halfWidthBins = (int)Math.ceil(halfWidth / bins.binWidth);
    int bin = bins.hitBin(w.getGF(Geom.offset(w.source, location, w.absoluteBearing)));
    double danger = 0;
    double[] dangers = getDangerTracker(w).getDangers();
    for (int i = Math.max(0, bin - halfWidthBins);
        i <= Math.min(bins.nBins - 1, bin + halfWidth); i++) {
      danger += dangers[i];
    }
    return danger * halfWidth / (halfWidthBins * bins.binWidth);
  }

  public static class ApproximateDangerLocation {
    public double danger;
    public int start;
    public int end;
  }

  public ApproximateDangerLocation getBestApproximateDanger(
      MovementWave w, double minDistance, double offset1, double offset2) {
    double width = FastTrig.atan(Geom.MAX_BOT_WIDTH / minDistance);
    width /= Math.min(w.maePrecise[0], w.maePrecise[1]);
    int widthBins = (int)Math.ceil(width / bins.binWidth);
    int startBin = bins.hitBin(w.getGF(offset1));
    int endBin = bins.hitBin(w.getGF(offset2));
    if (startBin > endBin) {
      int tmp = startBin;
      startBin = endBin;
      endBin = tmp;
    }
    startBin = Math.max(0, startBin - widthBins / 2);
    endBin = Math.min(bins.nBins - 1, endBin + widthBins / 2);
    double[] dangers = getDangerTracker(w).getDangers();
    double danger = 0;
    ApproximateDangerLocation loc = new ApproximateDangerLocation();
    loc.danger = Double.POSITIVE_INFINITY;
    for (int i = startBin; i <= endBin; i++) {
       danger += dangers[i];
       if (i >= startBin + widthBins) {
         danger -= dangers[i - widthBins];
       }
       if ((i >= startBin + widthBins || i == endBin) && danger < loc.danger) {
         loc.danger = danger;
         loc.end = i;
       }
    }
    loc.start = Math.max(0, loc.end - widthBins);
    loc.danger *= width / (widthBins * bins.binWidth);
    return loc;
  }

  public void paint() {
    for (Map.Entry<MovementWave, WaveDangerTracker> e : waveDangerTrackers.entrySet()) {
      MovementWave w = e.getKey();
      if (w.hasBullet && w.radius > 0) {
        double[] dangers = e.getValue().getDangers();
        double maxDanger = 0;
        for (int i = 0; i < dangers.length; i++) {
          maxDanger = Math.max(dangers[i], maxDanger);
        }
        for (int i = 0; i < dangers.length; i++) {
          maxDanger = Math.max(maxDanger, dangers[i]);
          double GF = (i + 0.5) * 2 / dangers.length - 1;
          int red = (int)(255 * dangers[i] / maxDanger);
          int green = (int)(255 * w.shadows[i]);
          Painter.THIS_TICK.addPoint(new Color(red, green, 0), Geom.project(
              w.source, w.absoluteBearing + w.getOffset(GF), w.radius - 7));
        }
      }
    }
  }

  public WaveDangerTracker getDangerTracker(MovementWave w) {
    WaveDangerTracker dangers = waveDangerTrackers.get(w);
    if (dangers == null) {
      dangers = new WaveDangerTracker(w);
      waveDangerTrackers.put(w, dangers);
    }
    return dangers;
  }

  public class WaveDangerTracker {
    private final MovementWave wave;
    private final Map<DangerEstimator, double[]> estimatorDangers;
    private double[] combinedDangers;
    private double[] shadowedDangers;

    private WaveDangerTracker(MovementWave wave) {
      this.wave = wave;
      estimatorDangers = new HashMap<>();
      combinedDangers = null;
      shadowedDangers = null;
    }

    public double[] getDangers() {
      if (shadowedDangers != null) {
        return shadowedDangers;
      }
      if (combinedDangers == null) {
        double[] unsmoothedDangers = new double[bins.nBins];
        /*for (int i = 0; i < unsmoothedDangers.length; i++) {
          unsmoothedDangers[i] += 0.1 / bins.nBins;
        }*/
        for (DangerEstimator estimator : activeEstimators) {
          Stopwatch.INSTANCE.start(estimator.isHitModel ? "Danger-Hits" : "Danger-Visits");
          double[] dangers = getEstimatorDangers(estimator);
          Stopwatch.INSTANCE.stop();
          for (int i = 0; i < dangers.length; i++) {
            unsmoothedDangers[i] += dangers[i] * estimator.weight;
          }
        }
        combinedDangers = new double[bins.nBins];
        int maxBinDiff = 20;
        for (int i = 0; i < bins.nBins; i++) {
          if (unsmoothedDangers[i] > 0) {
            for (int j = Math.max(0, i - maxBinDiff);
                j <= Math.min(bins.nBins - 1, i + maxBinDiff); j++) {
              double distance = Math.abs(i - j) * bins.binWidth;
              combinedDangers[j] += unsmoothedDangers[i] * Math.exp(Math.abs(distance) * -15);
            }
          }
        }
      }
      if (wave.shadows == null || antiHOT) {
        shadowedDangers = combinedDangers;
      } else {
        shadowedDangers = new double[bins.nBins];
        for (int i = 0; i < bins.nBins; i++) {
          shadowedDangers[i] = (1 - wave.shadows[i]) * combinedDangers[i];
        }
      }
      KUtils.normalize(shadowedDangers);
      return shadowedDangers;
    }

    public double[] getEstimatorDangers(DangerEstimator estimator) {
      if (estimatorDangers.containsKey(estimator)) {
        return estimatorDangers.get(estimator);
      }
      double[] dangers = estimator.model.getDangers(wave);
      KUtils.normalize(dangers);
      estimatorDangers.put(estimator, dangers);
      return dangers;
    }

    public void onUpdatedShadow() {
      shadowedDangers = null;
    }

    public void onActiveEstimatorsChanged() {
      shadowedDangers = combinedDangers = null;
    }

    public void onUpdatedEstimator(DangerEstimator estimator) {
      if (estimatorDangers.containsKey(estimator)) {
        shadowedDangers = combinedDangers = null;
        estimatorDangers.remove(estimator);
      }
    }
  }

  public void setActiveEstimators() {
    setEstimatorWeights(dangerEstimators);
    boolean activeEstimatorsChanged = false;
    for (DangerEstimator estimator : dangerEstimators) {
      if (estimator.isActive() && !activeEstimators.contains(estimator)) {
        activeEstimators.add(estimator);
        activeEstimatorsChanged = true;
        //System.out.println("Adding danger estimator " + estimator.name);
      }
      if (!estimator.isActive() && activeEstimators.contains(estimator)) {
        activeEstimators.remove(estimator);
        activeEstimatorsChanged = false;
        //System.out.println("Removing danger estimator " + estimator.name);
      }
    }
    setEstimatorWeights(activeEstimators);
    if (activeEstimatorsChanged) {
      for (WaveDangerTracker dangerTracker : waveDangerTrackers.values()) {
        dangerTracker.onActiveEstimatorsChanged();
      }
    }
  }

  private void setEstimatorWeights(List<DangerEstimator> estimators) {
    double total = 0;
    double priorTotal = 0;
    int minHits = Integer.MAX_VALUE;
    for (DangerEstimator estimator : estimators) {
      total += estimator.getUnnormalizedWeight();
      priorTotal += estimator.priorWeight;
      minHits = Math.min(minHits, estimator.hits);
    }
    double priorWeight = Math.max(PRIOR_STRENGTH / (minHits + PRIOR_STRENGTH), MIN_PRIOR_WEIGHT);
    for (DangerEstimator estimator : estimators) {
      estimator.weight = (priorWeight * estimator.priorWeight / priorTotal) +
          ((1 - priorWeight) * estimator.getUnnormalizedWeight() / total);
    }
  }

  public void printEstimatorWeight(DangerEstimator estimator, double total) {
    System.out.printf(" " + estimator.model.name + " %.3f\n",
        estimator.getUnnormalizedWeight() / (total / dangerEstimators.size()));
  }

  public void printEstimatorWeights() {
    double total = 0;
    for (DangerEstimator estimator : dangerEstimators) {
      total += estimator.getUnnormalizedWeight();
    }
    System.out.println("Inactive estimators:");
    for (DangerEstimator estimator : dangerEstimators) {
      if (!estimator.isActive()) {
        printEstimatorWeight(estimator, total);
      }
    }
    System.out.println("Active estimators:");
    for (DangerEstimator estimator : activeEstimators) {
      printEstimatorWeight(estimator, total);
    }
  }

  public class DangerEstimator {
    public final Model model;
    public final boolean isHitModel;
    public final double minHitRate;
    public final double maxHitRate;
    public final double priorWeight;

    private double emaDangerHit;
    private int hits;
    private double weight;

    public DangerEstimator(Model model, boolean isHitModel,
        double minHitRate, double maxHitRate, double priorWeight) {
      this.model = model;
      this.isHitModel = isHitModel;
      this.minHitRate = minHitRate;
      this.maxHitRate = maxHitRate;
      this.priorWeight = priorWeight;
    }

    public void updateWeightingStats(MovementWave w) {
      if (hits > 0) {
        WaveDangerTracker waveDangerTracker = getDangerTracker(w);
        double[] dangers = waveDangerTracker.getEstimatorDangers(this);
        double dangerEstimate = 1e-4;
        double halfWidth = FastTrig.atan(Geom.MAX_BOT_HALF_WIDTH / w.distance);
        halfWidth /= w.hitOffset > 0 ? w.maePrecise[1] : w.maePrecise[0];
        double[] GFRange = new double[] {Math.max(-0.999, w.hitGF() - halfWidth),
                                Math.min(0.999, w.hitGF() + halfWidth)};
        for (int bin = bins.hitBin(GFRange[0]); bin <= bins.hitBin(GFRange[1]); bin++) {
          dangerEstimate += bins.binWeight(GFRange, bin) * dangers[bin];
        }
        emaDangerHit = HIT_ROLLING * emaDangerHit + (1 - HIT_ROLLING) * dangerEstimate;
      }
      hits++;
    }

    public boolean isActive() {
      if (antiHOT) {
        return model.name.equals("HOT") || model.name.equals("Simple");
      }
      if (strategy.antiRam) {
        if (model.name.equals("Linear") || model.name.equals("Circular")) {
          return true;
        }
        if (!isHitModel) {
          return false;
        }
      }
      return hitRateTracker.hitRateInBounds(minHitRate, maxHitRate) ||
          weight > FORCE_ACTIVE_WEIGHT / dangerEstimators.size();
    }

    public double getUnnormalizedWeight() {
      if (hits <= 1) {
        return priorWeight;
      }
      return Math.pow(emaDangerHit / (1 - Math.pow(HIT_ROLLING, hits - 1)), WEIGHT_POWER);
    }
  }
}
