package kc.mega.movement;

import kc.mega.game.BattleField;
import kc.mega.game.BotState;
import kc.mega.game.GameState;
import kc.mega.game.PredictState;
import kc.mega.other.Painter;
import kc.mega.other.Strategy;
import kc.mega.utils.Geom;
import kc.mega.utils.KUtils;
import kc.mega.utils.Predict;
import kc.mega.wave.Wave;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jk.math.FastTrig;
import robocode.AdvancedRobot;

public class PathSurfer {
  private static final boolean TRUE_SURF = false;
  private static final double TARGET_DISTANCE = 600;

  private final AdvancedRobot bot;
  private final Strategy strategy;
  private final DangerComputer dangerComputer;
  private final Paths paths;

  private GameState gs;
  private List<MovementWave> waves;
  private List<Integer> currentPath;
  private List<SurfPlan> plans;
  private SurfPlan emptyPlan;

  public PathSurfer(AdvancedRobot bot, Strategy strategy, DangerComputer dangerComputer) {
    this.bot = bot;
    this.strategy = strategy;
    this.dangerComputer = dangerComputer;
    this.paths = new Paths();
  }

  public void surf(GameState gs, List<MovementWave> waves, boolean paint) {
    this.gs = gs;
    this.waves = waves;
    BotState myState = gs.getMyState(), enemyState = gs.getEnemyState();
    double distance = gs.getMyState().distance(enemyState);

    plans = new ArrayList<>();
    emptyPlan = new SurfPlan(myState.asPredictState());
    plans.add(emptyPlan);

    boolean preciseSurfSecond = waves.size() >= 2 &&
        waves.get(0).ticksUntilBreak + 4 >= waves.get(1).ticksUntilBreak;
    if (waves.size() > 0) {
      if (TRUE_SURF || strategy.antiRam) {
        trueSurfBranch();
      } else {
        pathSurfBranch();
        if (currentPath.size() > 1) {
          currentPath.remove(0);
          addIfNotDuplicate(currentPath, plans, emptyPlan.getSurfWave());
        }
      }
      setDangerMultipliers();
      if (preciseSurfSecond) {
        trueSurfBranch();
      } else {
        setApproximateDangers();
      }
    } else {
      noWaveBranch(KUtils.clip((int)(distance / 16), 1, 10));
      setDangerMultipliers();
    }

    SurfPlan bestPlan = getRankedPlans().get(0);
    currentPath = bestPlan.path;
    List<MoveInstruction> instructions = bestPlan.getInstructions();

    BotState futureState = gs.getMyState();
    List<BotState> futureStates = new ArrayList<>();
    for (int i = 0; i < instructions.size(); i++) {
      MoveInstruction instruction = instructions.get(i);
      futureState = futureState.getNextState(instruction.direction, instruction.turn);
      futureStates.add(futureState);
      if (paint) {
        Painter.THIS_TICK.addPoint(instruction.direction == 0 ? Color.gray.brighter() :
          (instruction.direction == 1 ? Color.yellow : Color.red), futureState.location, 2);
        if (i == instructions.size() - 1) {
          Painter.THIS_TICK.addBot(Color.white, futureState.location);
        }
      }
    }
    if (paint && waves.size() > 0 &&
        bestPlan.visitOffsetRanges.containsKey(waves.get(0))) {
      Wave surfWave = waves.get(0);
      if (surfWave.radius > 0) {
        double[] offsets = bestPlan.visitOffsetRanges.get(surfWave);
        surfWave.paintTick(Color.white, offsets[0]);
        surfWave.paintTick(Color.white, offsets[1]);
      }
      if (bestPlan.secondWaveLoc != null) {
        surfWave = waves.get(1);
        surfWave.paintTick(Color.gray, surfWave.getOffset(
            dangerComputer.getBins().midPoint[bestPlan.secondWaveLoc.start]));
        surfWave.paintTick(Color.gray, surfWave.getOffset(
            dangerComputer.getBins().midPoint[bestPlan.secondWaveLoc.end]));
      }
    }
    gs.setMyNextStates(futureStates);

    bot.setTurnRightRadians(instructions.get(0).turn);
    int direction = instructions.get(0).direction;
    bot.setAhead(direction == 0 ? (Math.random() - 0.5) / 1e12 :
      direction * Double.POSITIVE_INFINITY);
    if (strategy.ram) {
      double targetHeading = getTargetHeading(myState, bestPlan.path.get(0),
          myState.absoluteBearing(enemyState), distance, 0);
      double turn = FastTrig.normalRelativeAngle(targetHeading - bot.getHeadingRadians());
      bot.setMaxVelocity(Math.max(0, Math.abs(turn) < Math.PI / 9 ? 8 :
        8 - Math.abs(turn / (Math.PI / 6))));
    } else {
      bot.setMaxVelocity(8 - Math.random() / 1e12);
    }
  }

  public void addTrueSurfChildren(SurfPlan parent, List<SurfPlan> surfPlans, int ticks) {
    MovementWave surfWave = parent.getSurfWave();
    SurfPlan[] children = strategy.antiRam ?
        new SurfPlan[] {
            new SurfPlan(parent, Paths.FORWARD_PATH, 1),
            new SurfPlan(parent, Paths.BACKWARD_PATH, 1),
            new SurfPlan(parent, Paths.HALT_PATH, 1),
            new SurfPlan(parent, Paths.FORWARD_PATH, -1),
            new SurfPlan(parent, Paths.BACKWARD_PATH, -1),
            new SurfPlan(parent, Paths.HALT_PATH, -1)} :
         new SurfPlan[] {
            new SurfPlan(parent, Paths.FORWARD_PATH),
            new SurfPlan(parent, Paths.BACKWARD_PATH),
            new SurfPlan(parent, Paths.HALT_PATH)};
    for (SurfPlan child : children) {
      child.predict(surfWave, ticks, true);
      surfPlans.add(child);
    }
  }

  public void noWaveBranch(int ticks) {
    plans.clear();
    addTrueSurfChildren(emptyPlan, plans, ticks);
  }

  public void trueSurfBranch() {
    List<SurfPlan> newPlans = new ArrayList<>();
    for (SurfPlan parent : plans) {
      if (parent.getSurfWave() == null) {
        newPlans.add(parent);
      } else {
        addTrueSurfChildren(parent, newPlans, 1000);
      }
    }
    plans = newPlans;
  }

  public void pathSurfBranch() {
    List<SurfPlan> newPlans = new ArrayList<>();
    SurfPlan forward = new SurfPlan(emptyPlan, Paths.FORWARD_PATH);
    SurfPlan backward = new SurfPlan(emptyPlan, Paths.BACKWARD_PATH);
    MovementWave surfWave = emptyPlan.getSurfWave();
    forward.predict(surfWave, 1000, true);
    backward.predict(surfWave, 1000, true);
    newPlans.add(forward);
    newPlans.add(backward);

    List<Integer> bestForwardPath = getCandidatePath(surfWave, forward);
    List<Integer> bestBackwardPath = getCandidatePath(surfWave, backward);
    if (bestForwardPath != null) {
      addIfNotDuplicate(bestForwardPath, newPlans, surfWave);
    }
    if (bestBackwardPath != null) {
      addIfNotDuplicate(bestBackwardPath, newPlans, surfWave);
    }
    if (true) { //backwardPath == null || backwardPath == null) {
      addIfNotDuplicate(Paths.HALT_PATH, newPlans, surfWave);
    }
    plans = newPlans;
  }

  public List<Integer> getCandidatePath(MovementWave surfWave, SurfPlan plan) {
    if (plan.states.size() <= 2) {
      return null;
    }
    Point2D.Double bestPoint = null;
    double bestDanger = Double.MAX_VALUE;
    double bestDistance = 0;
    for (int i = 0; i < plan.states.size(); i++) {
      double danger = dangerComputer.getApproximatePointDanger(
          surfWave, plan.states.get(i).location);
      if (danger < bestDanger) {
        bestDanger = danger;
        bestDistance = plan.travelDistances.get(i);
        bestPoint = plan.states.get(i).location;
      }
    }
    double targetVelocity = Math.random() < 0.5 ? 0 : (Math.random() < 0.5 ? -8 : 8);
    List<Integer> path = paths.bestPath(plan.states.get(0).velocity,
        Math.max(1, surfWave.getTicksUntilPasses(bestPoint, gs.gameTime) - 1),
        bestDistance, targetVelocity);
    return path;
  }

  public void addIfNotDuplicate(List<Integer> path, List<SurfPlan> plans, MovementWave surfWave) {
    for (SurfPlan plan : plans) {
      int pathDirection = 0;
      int planDirection = 0;
      boolean different = false;
      for (int i = 0; i < Math.max(path.size(), plan.path.size()); i++) {
        if (i < path.size()) {
          pathDirection = path.get(i);
        }
        if (i < plan.path.size()) {
          planDirection = plan.path.get(i);
        }
        if (pathDirection != planDirection) {
          different = true;
          break;
        }
      }
      if (!different) {
        return;
      }
    }
    SurfPlan plan = new SurfPlan(emptyPlan, path);
    plan.predict(surfWave, 1000, true);
    plans.add(plan);
  }

  public void setDangerMultipliers() {
    for (SurfPlan plan : plans) {
      plan.setDangerMultiplier();
    }
  }

  public void setApproximateDangers() {
    double bestDanger = Double.POSITIVE_INFINITY;
    for (SurfPlan plan : getRankedPlans()) {
      if (plan.getDanger() > bestDanger) {
        break;
      }
      plan.setApproximateDanger();
      double danger = plan.getDanger();
      if (danger < bestDanger) {
        bestDanger = danger;
      }
    }
  }

  public List<SurfPlan> getRankedPlans() {
    for (SurfPlan plan : plans) {
      plan.setPreciseDangers();
    }
    Collections.sort(plans);
    return plans;
  }

  public static double getWaveWeight(MovementWave w) {
    return (1 + w.power) / Math.sqrt(2 + Math.max(1, w.ticksUntilBreak));
  }

  private class SurfPlan implements Comparable<SurfPlan> {
    public final List<PredictState> states;
    public final Map<MovementWave, double[]> visitOffsetRanges;
    public final List<Double> travelDistances;
    public final int antiRamEvadeSide;
    public List<Integer> path;
    public boolean hitWall;

    public double preciseDanger;
    public double approximateDanger;
    public double dangerMultiplier;
    public DangerComputer.ApproximateDangerLocation secondWaveLoc;

    public SurfPlan(PredictState state) {
      states = new ArrayList<>();
      states.add(state);
      travelDistances = new ArrayList<>();
      travelDistances.add(0.0);
      path = new ArrayList<>();
      visitOffsetRanges = new HashMap<>();
      hitWall = false;
      preciseDanger = approximateDanger = 0;
      dangerMultiplier = 1;
      antiRamEvadeSide = 1;
    }

    public SurfPlan(SurfPlan parent, List<Integer> extension) {
      this(parent, extension, 1);
    }

    public SurfPlan(SurfPlan parent, List<Integer> extension, int antiRamEvadeSide) {
      this.antiRamEvadeSide = antiRamEvadeSide;
      states = new ArrayList<>(parent.states);
      travelDistances = new ArrayList<>(parent.travelDistances);
      path = new ArrayList<>(parent.path);
      visitOffsetRanges = new HashMap<>();
      for (Map.Entry<MovementWave, double[]> e : parent.visitOffsetRanges.entrySet()) {
        this.visitOffsetRanges.put(e.getKey(), e.getValue().clone());
      }
      hitWall = parent.hitWall;
      preciseDanger = approximateDanger = 0;
      dangerMultiplier = parent.dangerMultiplier;
      path = new ArrayList<>(parent.path);
      path.addAll(extension);
    }

    private void predict(MovementWave surfWave, int maxTicks, boolean precise) {
      PredictState state = getEndState();
      double distanceTraveled = travelDistances.get(travelDistances.size() - 1);
      //Point2D.Double source = surfWave == null ? gs.getEnemyState().location : surfWave.source;
      Point2D.Double source = gs.getEnemyState().location;
      int tick = 0;
      int direction = 0;
      int nonzeroDirection = 0;
      boolean surfWavePassed = false;
      while (!surfWavePassed && tick <= maxTicks) {
        //Timer.INSTANCE.start("Next Tick Predict");
        tick = (int)(state.gameTime - states.get(0).gameTime);
        if (tick < path.size()) {
          direction = path.get(tick);
          if (direction != 0) {
            nonzeroDirection = direction;
          }
        } else {
          path.add(direction);
        }
        double absoluteBearing = Geom.absoluteBearing(state.location, source);
        double distance = state.location.distance(source);
        double targetHeading = getTargetHeading(
            state, nonzeroDirection == 0 ? KUtils.nonzeroSign(state.velocity) : nonzeroDirection,
                absoluteBearing, distance, antiRamEvadeSide);
        double turn = FastTrig.normalRelativeAngle(targetHeading - state.heading);
        double prevVel = state.velocity;
        state = state.getNextState(direction, turn);
        PredictState inFieldState = BattleField.INSTANCE.wallCollisionCheck(state, true);
        if (inFieldState != state) {
          state = inFieldState;
          if (Math.abs(prevVel) > 2) {
            hitWall = true;
          }
        }
        states.add(state);
        distanceTraveled += state.velocity;
        travelDistances.add(distanceTraveled);
        //Timer.INSTANCE.stop();

        //Timer.INSTANCE.start("Handle Waves");
        //TODO: ignore far-off waves, make this faster
        if (precise) {
          for (MovementWave w : waves) {
            Wave.Foam foam = w.getFoam(state);
            if (w == surfWave && foam.status >= Wave.WILL_PASS_THIS_TICK) {
              surfWavePassed = true;
            }
            if (foam.status == Wave.BREAKING || foam.status == Wave.WILL_PASS_THIS_TICK) {
              double[] offsetRange = visitOffsetRanges.get(w);
              if (offsetRange == null) {
                visitOffsetRanges.put(w, foam.hitOffsetRange);
              } else {
                Geom.updateOffsetRange(offsetRange, foam.hitOffsetRange);
              }
            }
          }
        } else {
          surfWavePassed = surfWave.source.distance(state.location) <
              (1 + state.gameTime - surfWave.fireTime) * surfWave.speed;
        }
        //Timer.INSTANCE.stop();
      }
    }

    public MovementWave getSurfWave() {
      PredictState state = getEndState();
      int minTicksUntilPasses = 1000;
      MovementWave surfWave = null;
      for (MovementWave w : waves) {
        int ticksUntilPasses = w.getTicksUntilPasses(state);
        if (ticksUntilPasses > 1 && ticksUntilPasses < minTicksUntilPasses) {
          minTicksUntilPasses = ticksUntilPasses;
          surfWave = w;
        }
      }
      return surfWave;
    }

    public PredictState getEndState() {
      return states.get(states.size() - 1);
    }

    public void setPreciseDangers() {
      preciseDanger = 0;
      for (Map.Entry<MovementWave, double[]> e : visitOffsetRanges.entrySet()) {
        preciseDanger += dangerComputer.getDanger(e.getKey(), e.getValue()) *
            getWaveWeight(e.getKey());
      }
    }

    public void setApproximateDanger() {
      MovementWave surfWave = getSurfWave();
      if (surfWave == null) {
        approximateDanger = 0;
        return;
      }
      SurfPlan forward = new SurfPlan(this, Paths.FORWARD_PATH);
      SurfPlan backward = new SurfPlan(this, Paths.BACKWARD_PATH);
      forward.predict(surfWave, 1000, false);
      backward.predict(surfWave, 1000, false);
      secondWaveLoc = dangerComputer.getBestApproximateDanger(
          surfWave, Math.min(getEndState().location.distance(surfWave.source),
                    Math.min(forward.getEndState().location.distance(surfWave.source),
                             backward.getEndState().location.distance(surfWave.source))),
          Geom.offset(surfWave.source, forward.getEndState().location,
              surfWave.absoluteBearing),
          Geom.offset(surfWave.source, backward.getEndState().location,
              surfWave.absoluteBearing));
      approximateDanger = getWaveWeight(surfWave) * secondWaveLoc.danger;
    }

    public void setDangerMultiplier() {
      double distanceRatio = states.get(0).location.distance(gs.getEnemyState().location) /
          getEndState().location.distance(gs.getEnemyState().location);
      dangerMultiplier = Math.exp(distanceRatio * (strategy.ram ? -1.5 : 1.5));
      double bulletSpeed = KUtils.bulletSpeed(gs.lastEnemyBulletPower);
      dangerMultiplier *= Math.pow(Geom.wallRestriction(
          getEndState().location, gs.getEnemyState().location, bulletSpeed), 1);
      dangerMultiplier /= Math.pow(Geom.wallRestriction(
          gs.getEnemyState().location, getEndState().location, bulletSpeed), 2);

      if (strategy.antiRam) {
        dangerMultiplier /= Math.sqrt(16 + BattleField.INSTANCE.wallDistance(
            getEndState().location));
        PredictState firstState = states.get(Math.min(states.size() - 1, 1));
        dangerMultiplier /= 16 + Math.abs(firstState.velocity * Math.sin(
            firstState.heading - gs.getEnemyState().absoluteBearing(firstState)));
      }
    }

    public double getDanger() {
      return (
          preciseDanger + approximateDanger +
          (strategy.antiRam ? 0.1 : 0.01) +
          strategy.getAntiMirrorDanger(getEndState(), preciseDanger + approximateDanger) +
          (hitWall ? 0.2 : 0) +
          ((strategy.antiRam || strategy.ram) ? 0 :
            Math.max(0, 1 - getEndState().distance(gs.getEnemyState()) / 300))
          ) * dangerMultiplier;
    }

    public List<MoveInstruction> getInstructions() {
      List<MoveInstruction> instructions = new ArrayList<>();
      for (int i = 0; i < states.size() - 1; i++) {
        instructions.add(new MoveInstruction(path.get(i),
            FastTrig.normalRelativeAngle(states.get(i + 1).heading - states.get(i).heading)));
      }
      return instructions;
    }

    @Override
    public int compareTo(SurfPlan other){
      return (int)Math.signum(getDanger() - other.getDanger());
    }
  }

  public static class MoveInstruction {
    public final int direction;
    public final double turn;

    public MoveInstruction(int direction, double turn) {
      this.direction = direction;
      this.turn = turn;
    }
  }

  public double getTargetHeading(
      PredictState state,
      int direction,
      double absoluteBearing,
      double distance,
      int antiRamEvadeSide) {
    //Timer.INSTANCE.start("Heading");
    absoluteBearing += Math.PI;
    double nextVelocity = Predict.getNextVelocity(state.velocity, direction);
    direction = nextVelocity == 0 ? direction : (int)Math.signum(nextVelocity);
    int orbitDirection = KUtils.orbitDirection(absoluteBearing, state.heading, direction);

    double targetHeading;
    if (strategy.antiRam) {
      targetHeading = absoluteBearing + antiRamEvadeSide * Math.max(0.7, 1 - distance / 300);
      if (direction == -1) {
        targetHeading += Math.PI;
      }
    } else if (strategy.ram) {
      PredictState enemyFutureState = gs.getEnemyState();
      for (int i = 0; i < Math.min(10, distance / 10); i++) {
        enemyFutureState = enemyFutureState.getNextState(
            (int)Math.signum(gs.getEnemyState().velocity),
            FastTrig.normalRelativeAngle(gs.getEnemyState().heading -
                gs.enemyHistory.get(1).heading));
      }
      targetHeading = state.absoluteBearing(enemyFutureState);
      if (direction == -1) {
        targetHeading += Math.PI;
      }
      return targetHeading;
    } else {
      double attackAngle = (distance - TARGET_DISTANCE) / (waves.size() == 0 ?  400 : 600);
      direction = nextVelocity == 0 ? direction : (int)Math.signum(nextVelocity);
      targetHeading = absoluteBearing + orbitDirection * (attackAngle + direction * Math.PI / 2);
    }

    double turn = FastTrig.normalRelativeAngle(targetHeading - state.heading);
    double nextHeading = state.heading + Predict.getTurnIncrement(
        turn, state.velocity);
    if (strategy.walkingStickSmooth) {
      // staying away from walls makes simple targeters with wall collision checks easier to predict
      targetHeading = BattleField.INSTANCE.walkingStickSmooth(
          state.location, targetHeading, direction, orbitDirection, 150, 25);
      turn = FastTrig.normalRelativeAngle(targetHeading - state.heading);
      nextHeading = state.heading + Predict.getTurnIncrement(turn, state.velocity);
    }
    double offset = direction == 1 ? 0 : Math.PI;
    return BattleField.INSTANCE.fancyStickSmooth(
        FastTrig.normalAbsoluteAngle(nextHeading + offset),
        Math.abs(nextVelocity), state.location.x, state.location.y,
        orbitDirection) + offset;
  }
}
