package kc.mega.wave;

import kc.mega.game.BotState;
import kc.mega.game.PredictState;
import kc.mega.learning.GFBins;
import kc.mega.other.Painter;
import kc.mega.utils.Geom;
import kc.mega.utils.KUtils;
import kc.mega.utils.Predict;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

import jk.math.FastTrig;

public class Wave {
  public static final int MIDAIR = 0, BREAKING = 1, WILL_PASS_THIS_TICK = 2, PASSED = 3;

  public final Point2D.Double source;
  public final long fireTime;
  public final double power;
  public final double speed;
  public final boolean doPaint;

  public final double absoluteBearing;
  public final int orbitDirection;
  public final double[] maePrecise;

  public int ticksUntilBreak;
  public int ticksUntilPass;
  public double radius;
  public boolean hasBullet;
  public boolean didHit;
  public boolean didCollide;

  public GFBins bins;
  public double hitOffset = 1000;
  public double[] aggHitOffsetRange = new double[] {1000, -1000};
  public double[] shadows;

  public static class Builder {
    public GFBins bins;
    public List<BotState> myHistory;
    public List<BotState> enemyHistory;
    public double power;
    public boolean hasBullet = true;
    public boolean doPaint = false;

    public Builder bins(GFBins val) {bins = val; return this;}
    public Builder myHistory(List<BotState> val) {myHistory = val; return this;}
    public Builder enemyHistory(List<BotState> val) {enemyHistory = val; return this;}
    public Builder power(double val) {power = val; return this;}
    public Builder hasBullet(boolean val) {hasBullet = val; return this;}
    public Builder doPaint(boolean val) {doPaint = val; return this;}
    public Wave build() {return new Wave(this);}
  }

  public Wave(Builder builder) {
    BotState myState = builder.myHistory.get(0);
    this.source = myState.location;
    this.fireTime = myState.gameTime;
    this.power = builder.power;
    this.speed = KUtils.bulletSpeed(builder.power);
    this.doPaint = builder.doPaint;
    this.hasBullet = builder.hasBullet;
    this.didHit = false;
    this.didCollide = false;
    this.bins = builder.bins;

    BotState enemyState = builder.enemyHistory.get(1);
    Point2D.Double myNextLocation = builder.myHistory.get(0).location;
    int moveDirection = 0;
    for (int i = 1; i < builder.enemyHistory.size(); i++) {
      if (moveDirection == 0) {
        moveDirection = (int)Math.signum(builder.enemyHistory.get(i).velocity);
      } else {
        break;
      }
    }
    absoluteBearing = Geom.absoluteBearing(myNextLocation, enemyState.location);
    orbitDirection = KUtils.orbitDirection(
        absoluteBearing, enemyState.heading, moveDirection);
    maePrecise = Predict.preciseMaxEscapeAngles(myNextLocation, enemyState, speed, true);
    if (bins != null) {
      shadows = new double[bins.nBins];
    }
  }

  public int update(BotState enemyState) {
    ticksUntilBreak = getTicksUntilBreak(enemyState);
    ticksUntilPass = getTicksUntilPasses(enemyState);
    radius = (enemyState.gameTime - fireTime) * speed;

    Foam foam = getFoam(enemyState);
    if (foam.status == BREAKING || foam.status == WILL_PASS_THIS_TICK) {
      Geom.updateOffsetRange(aggHitOffsetRange, foam.hitOffsetRange);
    }
    if (foam.status >= WILL_PASS_THIS_TICK) {
      if (hitOffset > 100) {  // TODO: this > 100 check is sort of a hack
        hitOffset = (aggHitOffsetRange[0] + aggHitOffsetRange[1]) / 2;
      }
    }

    if (doPaint) {
      paint(enemyState, foam);
    }

    return foam.status;
  }

  public double getGF(double offset) {
    offset = FastTrig.normalRelativeAngle(offset);
    offset /= offset < 0 ? maePrecise[0] : maePrecise[1];
    offset *= orbitDirection;
    return KUtils.clip(offset, -0.99999, 0.99999);
  }

  public double[] getGFRange(double[] offsets) {
    double GF1 = getGF(offsets[0]);
    double GF2 = getGF(offsets[1]);
    return GF1 < GF2 ? new double[] {GF1, GF2} : new double[] {GF2, GF1};
  }

  public double hitGF() {
    return getGF(hitOffset);
  }

  public double[] hitGFRange() {
   return getGFRange(aggHitOffsetRange);
  }

  public double getOffset(double GF) {
    GF *= orbitDirection;
    GF *= GF < 0 ? maePrecise[0] : maePrecise[1];
    return GF;
  }

  public void paint(BotState enemyState, Foam foam) {
    if (hasBullet && radius > 0) {
      Color c = Color.gray;
      if (foam.status == PASSED) {
        c = Color.DARK_GRAY;
      } else if (foam.status > MIDAIR) {
        c = Color.gray.brighter();
      }
      Painter.THIS_TICK.addCircle(c, source, radius);
      Painter.THIS_TICK.addCircle(c, source, 3, true);
      Painter.THIS_TICK.addCircle(c, Geom.project(source, absoluteBearing, radius), 2, true);
      paintBearing(c, absoluteBearing, 5, radius);
    }
  }

  public void paintTick(Color color, double offset) {
    paintBearing(color, absoluteBearing + offset, radius - speed / 2, radius);
  }

  public void paintBearing(Color color, double bearing, double start, double stop) {
    Painter.THIS_TICK.addLine(color, Geom.project(source, bearing, start),
        Geom.project(source, bearing, stop));
  }

  private boolean[] addBulletShadowsForTime(Bullet b, long gameTime) {
    // TODO: faster check for no shadows changed
    boolean bulletPassed = true;
    double[] newShadows = null;
    for (int bulletOffset = -1; bulletOffset <= 0; bulletOffset++) {
      for (int waveOffset = -1; waveOffset <= 0; waveOffset++) {
        if (bulletOffset == -1 && waveOffset == -1) {
          continue;
        }
        if (gameTime - b.fireTime + bulletOffset <= 0 ||
            gameTime - fireTime + waveOffset <= 0) {
          bulletPassed = false;
          continue;
        }

        Point2D.Double bStart = b.getLocation(gameTime + bulletOffset);
        Point2D.Double bEnd = b.getLocation(gameTime + 1 + bulletOffset);
        double startDistance = bStart.distance(source);
        double endDistance = bEnd.distance(source);
        double startRadius = (gameTime - fireTime + waveOffset) * speed;
        double endRadius = (gameTime - fireTime + 1 + waveOffset) * speed;
        if (startDistance < startRadius || endDistance > startDistance) {
          continue;
        }
        bulletPassed = false;
        if (startDistance < startRadius || endDistance > endRadius) {
          continue;
        }

        Point2D.Double shadowStart, shadowEnd;
        if (startDistance <= endRadius) {
          shadowStart = bStart;
        } else {
          shadowStart = Geom.circleLineSegIntercept(bStart, bEnd, source, endRadius);
        }
        if (endDistance >= startRadius) {
          shadowEnd = bEnd;
        } else {
          shadowEnd = Geom.circleLineSegIntercept(bStart, bEnd, source, startRadius);
        }
        double[] GFRange = getGFRange(new double[] {
            Geom.offset(source, shadowStart, absoluteBearing),
            Geom.offset(source, shadowEnd, absoluteBearing),
        });
        if (GFRange[0] < 0.99 || GFRange[1] > -0.99) {
          if (newShadows == null) {
            newShadows = new double[bins.nBins];
          }
          for (int bin = bins.hitBin(GFRange[0]); bin <= bins.hitBin(GFRange[1]); bin++) {
            if (waveOffset == bulletOffset) {
              newShadows[bin] += (1 - newShadows[bin]) * bins.binWeight(GFRange, bin);
            } else {
              newShadows[bin] += 0.5 * bins.binWeight(GFRange, bin);
            }
          }
        }
      }
    }

    if (newShadows != null) {
      for (int bin = 0; bin < newShadows.length; bin++) {
        if (newShadows[bin] > 0) {
          shadows[bin] += (1 - shadows[bin]) * newShadows[bin];
        }
      }
    }

    return new boolean[] {bulletPassed, newShadows == null};
  }

  public boolean addBulletShadows(Bullet b, long gameTime) {
    boolean shadowsChanged = false;
    for (long time = gameTime; ; time++) {
      boolean[] result = addBulletShadowsForTime(b, time);
      shadowsChanged |= result[1];
      if (result[0]) {
        return shadowsChanged;
      }
    }
  }

  public Foam getFoam(PredictState state) {
    double radius = (state.gameTime - fireTime) * speed;
    double nextRadius = radius + speed;
    double distance = state.location.distance(source);

    if (radius <= 0) {
      return new Foam(MIDAIR, null);
    }
    if (nextRadius < distance - Geom.MAX_BOT_HALF_WIDTH) {
      return new Foam(MIDAIR, null);
    }
    if (radius > distance + Geom.MAX_BOT_HALF_WIDTH) {
      return new Foam(PASSED, null);
    }

    double[] radiusSq_nextRadiusSq = new double[] {
        KUtils.sqr(radius), KUtils.sqr(nextRadius)};
    double[] yBounds = new double[] {state.location.y - 18, state.location.y + 18};
    double[] xBounds = new double[] {state.location.x - 18, state.location.x + 18};
    ArrayList<Point2D.Double> intercepts = new ArrayList<>();
    double maxCornerDistanceSq = 0;

    for (int i = 0; i < 2; i++) {
      for (int j = 0; j < 2; j++) {
        Point2D.Double corner = new Point2D.Double(xBounds[i], yBounds[j]);
        double cornerDistance = corner.distanceSq(source);
        if (cornerDistance >= radiusSq_nextRadiusSq[0] &&
            cornerDistance <= radiusSq_nextRadiusSq[1]) {
          intercepts.add(corner);
        }
        if (cornerDistance > maxCornerDistanceSq) {
          maxCornerDistanceSq = cornerDistance;
        }
      }
    }

    if (radius > Math.sqrt(maxCornerDistanceSq)) {
      return new Foam(PASSED, null);
    }

    for (int beforeTick = 0; beforeTick < 2; beforeTick++) {
      double radiusSq = radiusSq_nextRadiusSq[beforeTick];
      for (double y : yBounds) {
        for (double x: Geom.getXIntercepts(source, radiusSq, y)) {
          if (KUtils.inBounds(x, xBounds)) {
            intercepts.add(new Point2D.Double(x, y));
          }
        }
      }
      for (double x : xBounds) {
        for (double y: Geom.getYIntercepts(source, radiusSq, x)) {
          if (KUtils.inBounds(y, yBounds)) {
            intercepts.add(new Point2D.Double(x, y));
          }
        }
      }
    }

    if (intercepts.isEmpty()) {
      return new Foam(MIDAIR, null);
    }

    double[] hitOffsetRange = new double[] {1000, -1000};
    for (Point2D.Double intercept : intercepts) {
      Geom.updateOffsetRange(hitOffsetRange, FastTrig.normalRelativeAngle(
          Geom.absoluteBearing(source, intercept) - absoluteBearing));
    }

    return new Foam(
        nextRadius > Math.sqrt(maxCornerDistanceSq) ? WILL_PASS_THIS_TICK: BREAKING,
        hitOffsetRange
    );
  }

  public int getTicksUntilBreak(PredictState state) {
    int closeTicks = (int)(Geom.getMinDistanceToBot(source, state.location) / speed);
    return closeTicks - (int)(state.gameTime - fireTime);
  }

  public int getTicksUntilPasses(PredictState state) {
    return getTicksUntilPasses(state.location, state.gameTime);
  }

  public int getTicksUntilPasses(Point2D.Double location, long time) {
    int passTicks = 1 + (int)(Geom.getMaxDistanceToBot(source, location) / speed);
    return passTicks - (int)(time - fireTime);
  }

  public static class Foam {
    public final int status;
    public final double[] hitOffsetRange;

    public Foam(int state, double[] hitOffsetRange) {
      this.status = state;
      this.hitOffsetRange = hitOffsetRange;
    }
  }

  public static class Bullet {
    public final Point2D.Double source;
    public final double heading;
    public final long fireTime;
    public final double dx;
    public final double dy;

    public Bullet(Point2D.Double source, double heading, double speed, long fireTime) {
      this.source = source;
      this.heading = heading;
      this.fireTime = fireTime;
      dx = Math.sin(heading) * speed;
      dy = Math.cos(heading) * speed;
    }

    public Point2D.Double getLocation(long gameTime) {
      return new Point2D.Double(source.x + dx * (gameTime - fireTime),
                                source.y + dy * (gameTime - fireTime));
    }
  }
}
