package kc.mega.gun;

import kc.mega.game.BattleField;
import kc.mega.game.GameState;
import kc.mega.other.HitRateTracker;
import kc.mega.utils.KUtils;

public class BulletPowerSelector {
  // Bullet powers found by Skilgannon that exploit a bug in BasicSurfer
  private static final double[] CANDIDATE_POWERS_ABS = new double[] {
      2.95,       1.95,             1.15, 0.95, 0.85,       0.65, 0.45, 0.35, 0.25, 0.15};
  private static final double[] CANDIDATE_POWERS = new double[] {
      2.95, 2.45, 1.95, 1.65, 1.45, 1.15, 0.95, 0.85, 0.75, 0.65, 0.45, 0.35, 0.25, 0.15, 0.1};
  private static final double FULL_POWER_DISTANCE = 120;
  private static final double MIN_ENERGY = 0.05;

  private static final double POINTS_FOR_WIN = 60;  // really 50, but we slightly favor survival
  private static final double BULLET_DAMAGE_BONUS = 0.2;

  private final HitRateTracker myHitRateTracker;
  private final HitRateTracker enemyHitRateTracker;
  private final double myEnergy;
  private final double enemyEnergy;
  private final BulletPower enemyPower;

  public BulletPowerSelector(GameState gs) {
    myHitRateTracker = gs.myHitRateTracker;
    enemyHitRateTracker = gs.enemyHitRateTracker;
    myEnergy = gs.getMyState().energy;
    enemyEnergy = gs.getEnemyState().energy;
    enemyPower = new BulletPower(Math.min(enemyEnergy,
        gs.lastEnemyBulletPower < 0 ? 1.95 : gs.lastEnemyBulletPower), enemyHitRateTracker);
  }

  public static double bestBulletPower(GameState gs) {
    BulletPowerSelector selector = new BulletPowerSelector(gs);
    boolean closeDistance = gs.getMyState().distance(gs.getEnemyState()) < FULL_POWER_DISTANCE;
    boolean antiBasicSurfer = gs.roundNum == 0 || selector.myHitRateTracker.getMaxHitRate() > 0.2;
    double myEnergy = gs.getMyState().energy;
    double enemyEnergy = gs.getEnemyState().energy;
    double minimumKillPower = 0.001 + (enemyEnergy / 4);
    if (enemyEnergy > 4) {
      minimumKillPower = 0.001 + ((enemyEnergy + 2) / 6);
    }

    double bestPower = 0;
    if (closeDistance) {
      bestPower = Math.nextAfter(2.95, -1);
    } else {
      double bestScore = Double.NEGATIVE_INFINITY;
      for (double power : (antiBasicSurfer ? CANDIDATE_POWERS_ABS : CANDIDATE_POWERS)) {
        double score = selector.scorePower(power);
        if (score > bestScore) {
          bestScore = score;
          bestPower = Math.nextAfter(power, -1);
        }
      }
    }

    if (!antiBasicSurfer) {
      bestPower = KUtils.clip(bestPower, 0.1, minimumKillPower);
      bestPower = Math.min(bestPower, myEnergy - MIN_ENERGY);
      if (closeDistance) {
        bestPower = Math.max(bestPower, 0.1);
      } else if (myEnergy < 5 && myEnergy > enemyEnergy) {
        bestPower = Math.min(myEnergy - enemyEnergy - 0.01, bestPower);
      }
    } else {
      for (int i = CANDIDATE_POWERS_ABS.length - 1; i >= 0; i--) {
        if (CANDIDATE_POWERS_ABS[i] > minimumKillPower) {
          bestPower = Math.min(Math.nextAfter(CANDIDATE_POWERS_ABS[i], -1), bestPower);
          break;
        }
      }
      for (double power : CANDIDATE_POWERS_ABS) {
        if (bestPower > myEnergy && power < myEnergy) {
          bestPower = Math.nextAfter(power, -1);
          break;
        }
      }
      if (bestPower > myEnergy ||
          (myEnergy < 5 && myEnergy > enemyEnergy && myEnergy - bestPower < enemyEnergy)) {
        bestPower = -1;
      }
    }

    return bestPower;
  }

  private double scorePower(double power) {
    BulletPower myPower = new BulletPower(power, myHitRateTracker);

    double ticksLeft = getTicksLeft(myPower, myPower.hitRate, enemyPower.hitRate);
    double myDamage = myPower.damage * myPower.hitRate * ticksLeft;
    double enemyDamage = enemyPower.damage * enemyPower.hitRate * ticksLeft;

    double a = myEnergy * myPower.damage + enemyEnergy * myPower.gain;
    double b = enemyEnergy * enemyPower.damage + myEnergy * enemyPower.gain;
    double lambda = (enemyEnergy * myPower.loss) - (myEnergy * enemyPower.loss) -
        (myPower.hitRate * a) + (enemyPower.hitRate * b);
    lambda /= a * a + b * b;
    double myNeededHR = lambda * a + myPower.hitRate;
    double enemyNeededHR = -lambda * b + enemyPower.hitRate;

    double winProb;
    if (myNeededHR <= 0 || enemyNeededHR >= 1) {
      winProb = 1;
    } else if (myNeededHR >= 1 || enemyNeededHR <= 0) {
      winProb = 0;
    } else {
      ticksLeft = getTicksLeft(myPower, myNeededHR, enemyNeededHR);
      winProb = Math.sqrt((1 - myPower.probAboveHR(myNeededHR, ticksLeft)) *
          enemyPower.probAboveHR(enemyNeededHR, ticksLeft));
    }

    return myDamage - enemyDamage +
           POINTS_FOR_WIN * winProb - POINTS_FOR_WIN * (1 - winProb) +
           BULLET_DAMAGE_BONUS * (myHitRateTracker.getDamageThisRound() + myDamage) * winProb -
           BULLET_DAMAGE_BONUS * (enemyHitRateTracker.getDamageThisRound() + enemyDamage) *
             (1 - winProb);
  }

  public double getTicksLeft(BulletPower myPower, double myHitRate, double enemyHitRate) {
    double myCombinedLoss = Math.max(
        enemyPower.damage * enemyHitRate - myPower.gain * myHitRate + myPower.loss, 0.0001);
    double enemyCombinedLoss = Math.max(
        myPower.damage * myHitRate - enemyPower.gain * enemyHitRate + enemyPower.loss, 0.0001);
    return Math.min(myEnergy / myCombinedLoss, enemyEnergy / enemyCombinedLoss);
  }

  public static class BulletPower {
    public final int cooldown;
    public final double damage;
    public final double gain;
    public final double loss;
    public final double hitRate;

    public BulletPower(double power, HitRateTracker hitRateTracker) {
      cooldown = (int)((1 + power / 5) / BattleField.INSTANCE.getGunCoolingRate());
      damage = KUtils.bulletDamage(power) / cooldown;
      gain = 3 * power / cooldown;
      loss = power / cooldown;
      hitRate = hitRateTracker.estimateHitRate(power);
    }

    public double probAboveHR(double targetHR, double ticks) {
      double shots = ticks / cooldown;
      double expectedHits = hitRate * shots;
      double variance = hitRate * (1 - hitRate) * shots;
      double targetHits = targetHR * shots;
      return phi((targetHits - expectedHits) / Math.sqrt(variance));
    }
  }

  public static double erf(double z) {
    double t = 1.0 / (1.0 + 0.5 * Math.abs(z));
    double ans = 1 - t * Math.exp(-z * z - 1.26551223 +
        t * (1.00002368 + t * (0.37409196 + t * (0.09678418 + t * (-0.18628806 + t * (0.27886807 +
            t * (-1.13520398 + t * (1.48851587 + t * (-0.82215223 + t * (0.17087277))))))))));
    return z > 0 ? ans : -ans;
  }

  public static double phi(double x) {
    return (1.0 + erf(x / Math.sqrt(2.0))) / 2.0;
  }
}
